'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var server = require('react-dom/server');
var Babel = require('@babel/standalone');
var merge = _interopDefault(require('deepmerge'));
var memoize = _interopDefault(require('fast-memoize'));
var React = require('react');
var React__default = _interopDefault(React);
var PropTypes = _interopDefault(require('prop-types'));

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

function _objectWithoutProperties(source, excluded) {
  if (source == null) return {};

  var target = _objectWithoutPropertiesLoose(source, excluded);

  var key, i;

  if (Object.getOwnPropertySymbols) {
    var sourceSymbolKeys = Object.getOwnPropertySymbols(source);

    for (i = 0; i < sourceSymbolKeys.length; i++) {
      key = sourceSymbolKeys[i];
      if (excluded.indexOf(key) >= 0) continue;
      if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue;
      target[key] = source[key];
    }
  }

  return target;
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

var exportToIIFEPlugin = function exportToIIFEPlugin(_ref) {
  var t = _ref.types;
  return {
    visitor: {
      ExportDefaultDeclaration: function ExportDefaultDeclaration(path) {
        var program = path.find(function (parentPath) {
          return parentPath.isProgram();
        });
        var declaration = path.node.declaration;

        if (t.isClassDeclaration(declaration)) {
          path.replaceWith(declaration);
          program.pushContainer("body", t.ReturnStatement(t.identifier(declaration.id.name)));
          return;
        }

        program.pushContainer("body", t.ReturnStatement(t.identifier(declaration.name)));
        path.remove();
      },
      Program: {
        exit: function exit(path) {
          path.stop();
        }
      }
    }
  };
};

var resolverId = "__resolverFn__";

var createVariableDeclaration = function createVariableDeclaration(t, specifier, source) {
  return t.variableDeclaration("const", [t.variableDeclarator(t.identifier(specifier.local.name), t.callExpression(t.identifier(resolverId), [t.stringLiteral(source.value)]))]);
};

var createDestructedDeclaration = function createDestructedDeclaration(t, specifier, source) {
  return t.variableDeclaration("const", [t.variableDeclarator(t.objectPattern([t.objectProperty(t.identifier(specifier.imported.name), t.identifier(specifier.local.name), false, true)]), t.callExpression(t.identifier(resolverId), [t.stringLiteral(source.value)]))]);
};

var importResolverPlugin = function importResolverPlugin(_ref) {
  var t = _ref.types;
  return {
    visitor: {
      ImportDeclaration: function ImportDeclaration(path) {
        var replacements = path.node.specifiers.map(function (specifier) {
          if (t.isImportDefaultSpecifier(specifier) || t.isImportNamespaceSpecifier(specifier)) {
            return createVariableDeclaration(t, specifier, path.node.source);
          }

          if (t.isImportSpecifier(specifier)) {
            return createDestructedDeclaration(t, specifier, path.node.source);
          }

          return null;
        });
        path.replaceWithMultiple(replacements);
      }
    }
  };
};

Babel.registerPlugin("export-to-iife", exportToIIFEPlugin);
Babel.registerPlugin("import-resolver", importResolverPlugin);
/**
 * Transform passed ES6 code to the evaluable code using plugins and a passed config.
 *
 * @param {Object} babelConfig A config for Babel
 * @param {String} source A string that contains the source code
 *
 * @return {String}
 */

var transformSource = function transformSource(babelConfig, source) {
  var config = merge({
    plugins: ["import-resolver", "export-to-iife"],
    presets: ["react"]
  }, babelConfig);

  var _Babel$transform = Babel.transform(source, config),
      code = _Babel$transform.code;

  return code;
};

var transformSource$1 = memoize(transformSource);

var createFunction = memoize(function (source) {
  return new Function(resolverId, source);
});
/**
 * Evaluates passed source code, uses passed resolver to resolve imports.
 *
 * @param {String} source A string that contains the source code
 * @param {Function} resolver A function for the imports resolution
 */

var evalSource = memoize(function (source, resolver) {
  var evalWithResolver = createFunction(source);
  return evalWithResolver(resolver);
});
/**
 * Creates a valid React Component from source.
 *
 * @param {Object} babelConfig A config for Babel
 * @param {Function} resolver A function for the imports resolution
 * @param {Object} resolverContext A context of resolver, will be passed as second argument to it
 * @param {String} source A string that contains the source code
 */

var createComponentFromSource = function createComponentFromSource(babelConfig, resolver, resolverContext, source) {
  var transformed = transformSource$1(babelConfig, source);

  var resolverWithContext = function resolverWithContext(importName) {
    return resolver(importName, resolverContext);
  };

  return evalSource(transformed, resolverWithContext);
};

var FallbackRender = function FallbackRender(_ref) {
  var children = _ref.children;
  var keys = Object.keys(children).reverse();
  var found = keys.find(function (key) {
    return children[key] !== undefined;
  });
  return found ? children[found] : null;
};

FallbackRender.propTypes = {
  children: PropTypes.objectOf(PropTypes.element).isRequired
};

var SafeRender =
/*#__PURE__*/
function (_Component) {
  _inherits(SafeRender, _Component);

  function SafeRender() {
    var _getPrototypeOf2;

    var _this;

    _classCallCheck(this, SafeRender);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    _this = _possibleConstructorReturn(this, (_getPrototypeOf2 = _getPrototypeOf(SafeRender)).call.apply(_getPrototypeOf2, [this].concat(args)));

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "state", {
      // eslint-disable-next-line
      cycleId: 0,
      failed: false
    });

    return _this;
  }

  _createClass(SafeRender, [{
    key: "componentDidCatch",
    value: function componentDidCatch() {
      var _this$props = this.props,
          cycleId = _this$props.cycleId,
          onError = _this$props.onError;
      onError(cycleId);
      this.setState({
        failed: true
      });
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props2 = this.props,
          children = _this$props2.children,
          prevChildren = _this$props2.prevChildren;
      var failed = this.state.failed;
      return failed ? React__default.createElement(FallbackRender, null, prevChildren) : children;
    }
  }], [{
    key: "getDerivedStateFromProps",
    value: function getDerivedStateFromProps(props, state) {
      return {
        cycleId: props.cycleId,
        failed: props.cycleId === state.cycleId ? state.failed : false
      };
    }
  }]);

  return SafeRender;
}(React.Component);

_defineProperty(SafeRender, "propTypes", {
  children: PropTypes.node.isRequired,
  cycleId: PropTypes.oneOfType([PropTypes.number, PropTypes.string]).isRequired,
  onError: PropTypes.func.isRequired,
  prevChildren: PropTypes.objectOf(PropTypes.element)
});

_defineProperty(SafeRender, "defaultProps", {
  prevChildren: {}
});

var _createContext = React.createContext("SourceRender.Element"),
    Consumer = _createContext.Consumer,
    Provider = _createContext.Provider;

var SourceRender =
/*#__PURE__*/
function (_Component) {
  _inherits(SourceRender, _Component);

  function SourceRender() {
    var _getPrototypeOf2;

    var _this;

    _classCallCheck(this, SourceRender);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    _this = _possibleConstructorReturn(this, (_getPrototypeOf2 = _getPrototypeOf(SourceRender)).call.apply(_getPrototypeOf2, [this].concat(args)));

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "htmlMarkup", "");

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "currentElement", null);

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "renderedElements", {
      0: null
      /** Stores an incremented value of the render cycle. */

    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "renderCycleId", 0);

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "handleError", function (cycleId) {
      _this.renderedElements[cycleId] = undefined;
    });

    return _this;
  }

  _createClass(SourceRender, [{
    key: "render",
    value: function render() {
      var _this$props = this.props,
          babelConfig = _this$props.babelConfig,
          children = _this$props.children,
          renderHtml = _this$props.renderHtml,
          resolver = _this$props.resolver,
          resolverContext = _this$props.resolverContext,
          source = _this$props.source,
          rest = _objectWithoutProperties(_this$props, ["babelConfig", "children", "renderHtml", "resolver", "resolverContext", "source"]);

      try {
        var ComponentFromSource = createComponentFromSource(babelConfig, resolver, resolverContext, source);
        this.renderCycleId += 1;
        this.currentElement = React__default.createElement(SafeRender, {
          cycleId: this.renderCycleId,
          onError: this.handleError,
          prevChildren: this.renderedElements
        }, React__default.createElement(ComponentFromSource, rest));
        this.htmlMarkup = renderHtml ? server.renderToStaticMarkup(this.currentElement) : "";
        this.renderedElements[this.renderCycleId] = this.currentElement;
        this.error = undefined;
      } catch (e) {
        this.error = e;
      }

      return React__default.createElement(Provider, {
        value: {
          error: this.error,
          element: this.currentElement,
          markup: this.htmlMarkup
        }
      }, children);
    }
  }]);

  return SourceRender;
}(React.Component);

_defineProperty(SourceRender, "propTypes", {
  /** A config for Babel. */
  babelConfig: PropTypes.object,
  children: PropTypes.node.isRequired,

  /**
   * An option that controls rendering of HTML with ReactDOM server, it allows to omit
   * rendering when you're using portals.
   */
  renderHtml: PropTypes.bool,

  /**
   * A function for the imports resolution.
   *
   * @param {String} importPath
   */
  resolver: PropTypes.func.isRequired,

  /** An object that will be passed additionally to resolver function. */
  resolverContext: PropTypes.object,

  /** A string that contains the source code. */
  source: PropTypes.string.isRequired
});

_defineProperty(SourceRender, "defaultProps", {
  babelConfig: {},
  renderHtml: true,
  resolverContext: {}
  /** Stores an HTML markup for the current element. */

});

SourceRender.Consumer = Consumer;

module.exports = SourceRender;
